﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Sockets;
using System.Text;
using FunnyCommon.Remoting.Interfaces;
using FunnyCommon.Remoting.Protocol;

namespace FunnyCommon.Remoting.Networking
{
    public class Client : IRemoteGpoutManager
    {
        private class BulkCommand : IDisposable
        {
            private readonly Client _owner;
            private List<Command> _list;

            public BulkCommand(Client owner, int initalCapacity)
            {
                _owner = owner;
                _list = new List<Command>(initalCapacity);
            }

            public void PushCommand(Command command)
            {
                _list.Add(command);

                if (_list.Count != 65535) return;

                _owner.SendCommandBulk(_list.ToArray());
                _list = new List<Command>();
            }

            public void Dispose()
            {
                _owner._bulk = null;
                _owner.SendCommandBulk(_list.ToArray());
            }
        }

        private Socket _sck;
        private NetworkStream _ns;
        private BulkCommand _bulk;
        public TextWriter Log { get; set; }

        public void Connect(string host, int port)
        {
            Disconnect();
            _sck = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            _sck.Connect(host, port);
            _ns = new NetworkStream(_sck);
            _ns.WriteByte(Consts.Version);
        }

        public void Disconnect()
        {
            _ns?.Close();
            _sck?.Close();
        }

        public Response SendCommand(Command command)
        {
            if (_bulk != null)
            {
                _bulk.PushCommand(command);
                return null;
            }

            using (var br = new BinaryReader(_ns, Encoding.UTF8, true))
            using (var bw = new BinaryWriter(_ns, Encoding.UTF8, true))
            {
                bw.Write((byte)1); //count L
                bw.Write((byte)0); //count H

                bw.Write(command.ToBytes());

                ushort count = br.ReadByte(); //dummy count L
                count = (ushort)(br.ReadByte() * 0xFF); //dummy count H

                return Response.ReadFromStream(_ns);
            }
        }

        public Response[] SendCommandBulk(Command[] commands)
        {
            using (var br = new BinaryReader(_ns, Encoding.UTF8, true))
            using (var bw = new BinaryWriter(_ns, Encoding.UTF8, true))
            {
                bw.Write((byte)commands.Length); //L
                bw.Write((byte)(commands.Length >> 8)); //H

                foreach (var command in commands)
                {
                    bw.Write(command.ToBytes());
                }

                var lenL = br.ReadByte();
                var lenH = br.ReadByte();

                var len = (ushort) (lenL + (ushort)(lenH * 256));

                var r = new Response[len];
                for (var i = 0; i < len; i++)
                {
                    r[i] = Response.ReadFromStream(_ns);
                }
                return r;
            }
        }

        public IDisposable CreateBulk(int estimateMaxCount)
        {
            return _bulk = new BulkCommand(this, estimateMaxCount);
        }

        public bool? Allocate(byte pin)
        {
            var r = SendCommand(new Command { Type = CommandType.AllocateGpio, Value = pin });
            if (r == null) return false;
            Log?.Write($"{nameof(Allocate)} {pin} - Success:{r.Success}, Message:{r.Message}");
            return r.Success;
        }

        public bool Free(byte pin)
        {
            var r = SendCommand(new Command { Type = CommandType.FreeGpio, Value = pin });
            if (r == null) return false;
            Log?.Write($"{nameof(Free)} {pin} - Success:{r.Success}, Message:{r.Message}");
            return r.Success;
        }

        public void WritePin(byte pin, bool height)
        {
            var r = SendCommand(new Command { Type = height ? CommandType.SetHeight : CommandType.SetLow, Value = pin });
            if (r == null) return;
            Log?.Write($"{nameof(WritePin)} {pin},{height} - Success:{r.Success}, Message:{r.Message}");
            if (!r.Success) throw new IOException(r.Message);
        }

        public void Dispose()
        {
            Disconnect();
        }
    }
}