﻿using System;
using System.Diagnostics;
using System.Runtime.CompilerServices;

namespace OledSsd1306
{
    public class OledSsd1306Driver
    {
        public const int I2CAddress = 0x3C;
        private const int Width = 128;
        private const int Height = 32;

        private readonly Raspiic _bus;
        private readonly byte[] _buffer;

        private readonly Stopwatch _sw = new Stopwatch();

        public OledSsd1306Driver()
        {
            _bus = new Raspiic(I2CAddress);
            _buffer = new byte[1 + Height * Width / 8];
            _buffer[0] = 0x40;
        }

        private void SendCommand(params byte[] cmds)
        {
            foreach (var cmd in cmds) _bus.Write(0, cmd);
        }

        public void Init()
        {
            _bus.Init();

            //REF:github.com/iliapenev/ssd1306_i2c
            SendCommand(0xAE); //DISPLAYOFF
            SendCommand(0xD5, 0x80); //SETDISPLAYCLOCKDIV ,the suggested ratio 0x80
            SendCommand(0xA8, Height - 1); //SETMULTIPLEX
            SendCommand(0xD3, 0x00); // SETDISPLAYOFFSET,no offset
            SendCommand(0x40 | 0x00); //SETSTARTLINE , #0
            SendCommand(0x8D, 0x14); //CHARGEPUMP,SWITCHCAPVCC
            SendCommand(0x20, 0x00); //MEMORYMODE,0x0 act like ks0108
            SendCommand(0xA0 | 0x01);//SEGREMAP
            SendCommand(0xC8); //COMSCANDEC
            SendCommand(0xDA, 0x02); //SETCOMPINS
            SendCommand(0x81, 0x8F); //SETCONTRAST
            SendCommand(0xD9, 0xF1); //SETPRECHARGE
            SendCommand(0xDB, 0x40);//SETVCOMDETECT
            SendCommand(0xA4, 0xA6); //DISPLAYALLON_RESUME,NORMALDISPLAY
            SendCommand(0x2E, 0xAF); //DEACTIVATE_SCROLL,DISPLAYON
        }

        public void Off()
        {
            SendCommand(0xAE); //DISPLAYOFF
        }

        public void On()
        {
            SendCommand(0xAF); //DISPLAYON
        }

        public void Enable(bool on)
        {
            if (on) On();
            else Off();
        }

        public void SetContrast(byte contrast, byte precharge)//, byte comdetect = 0x40)
        {
            SendCommand(0xD9, precharge); //SETPRECHARGE,0xF1 default, to lower the contrast, put 1-1F
            SendCommand(0x81, contrast); //SETCONTRAST, 0-255
            //SendCommand(0xDB, comdetect); //SETVCOMDETECT, (additionally needed to lower the contrast)
            //                              //0x40 default, to lower the contrast, put 0

            //SendCommand(0xA4, 0xA6, 0xAF);//DISPLAYALLON_RESUME,NORMALDISPLAY,DISPLAYON
        }

        public void SetContrast(int level)
        {
            if (level < 0 || level > 511) throw new ArgumentOutOfRangeException(nameof(level));
            if (level < 256) SetContrast((byte)level, 0);
            else SetContrast((byte)(level - 256), 1);
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public void SetPixel(int x, int y, bool bw)
        {
            //REF:github.com/iliapenev/ssd1306_i2c
            switch (bw)
            {
                case true:
                    _buffer[1 + x + y / 8 * Width] |= (byte)(1 << (y & 7));
                    break;

                case false:
                    _buffer[1 + x + y / 8 * Width] &= (byte)(~(1 << (y & 7)));
                    break;
            }
        }

        public void Update()
        {
            _sw.Restart();
            SendCommand(0x21, 0, Width - 1); //COLUMNADDR
            SendCommand(0x22, 0, 3); //PAGEADDR


            _bus.Write(_buffer);

            _sw.Stop();
            Console.WriteLine($"#DEBUG SSD1306 Flush ms:{_sw.ElapsedMilliseconds}");
        }

        public void Clear(bool update = true)
        {
            Array.Clear(_buffer, 1, _buffer.Length-1);
            if (update) Update();
        }

        public void Fill(bool update = true)
        {
            var bufferLength = _buffer.Length;
            for (var index = 1; index < bufferLength; ++index) _buffer[index] = 0xFF;
            if (update) Update();
        }
    }
}